/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of the program kresilience                           */
/*                                                                           */
/*    Copyright (C) 2020-2021 Andreas Schmitt, Marc Pfetsch                  */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP --- Solving Constraint Integer Programs                  */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   sepa_2dConvexCube.c
 * @brief  assume a 2d cubic function which is indefinite but edge convex together with indicator structure
 * @author Andreas Schmitt
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <assert.h>

#include "sepa_2dConvexCube.h"
#include "min2dcube.h"


#define SEPA_NAME              "convexcube"
#define SEPA_DESC              "edge convex cubic function separator"
#define SEPA_PRIORITY               110
#define SEPA_FREQ                     1
#define SEPA_MAXBOUNDDIST           1.0
#define SEPA_USESSUBSCIP          FALSE      /**< does the separator use a secondary SCIP instance? */
#define SEPA_DELAY                FALSE      /**< should separation method be delayed, if other separators found cuts? */


/*
 * Data structures
 */

/** separator data */
struct SCIP_SepaData
{
   int                   neqs;               /**< number of inequalities/sets we consider */
   int                   sizeeqs;            /**< allocated size of all arrays in this struct */
   SCIP_VAR**            varsT;              /**< variables t */
   SCIP_VAR**            varsX;              /**< variables x */
   SCIP_VAR**            varsY;              /**< variables y */
   SCIP_VAR**            varsZ;              /**< indicator variables */
   SCIP_Real*            coefsXXX;           /**< coefficient of functions */
   SCIP_Real*            coefsXXY;           /**< coefficient of functions */
   SCIP_Real*            coefsXYY;           /**< coefficient of functions */
   SCIP_Real*            coefsYYY;           /**< coefficient of functions */
   SCIP_Real*            coefsXX;            /**< coefficient of functions */
   SCIP_Real*            coefsXY;            /**< coefficient of functions */
   SCIP_Real*            coefsYY;            /**< coefficient of functions */
   SCIP_Real*            coefsX;             /**< coefficient of functions */
   SCIP_Real*            coefsY;             /**< coefficient of functions */
   SCIP_Real*            coefsC;             /**< coefficient of functions */
};


/*
 * Local methods
 */

/* checks whether the 2 dimensional cubic function is convex in x for for x in [lbx, ubx] for all y in [lby, uby] */
static
SCIP_Bool isConvexInXOverBox(
   SCIP_Real             lbx,                /**< lower bound of x */
   SCIP_Real             ubx,                /**< upper bound of y */
   SCIP_Real             lby,                /**< lower bound of y */
   SCIP_Real             uby,                /**< upper bound of y */
   SCIP_Real             cxxx,               /**< function coefficient */
   SCIP_Real             cxxy,               /**< function coefficient */
   SCIP_Real             cxyy,               /**< function coefficient */
   SCIP_Real             cyyy,               /**< function coefficient */
   SCIP_Real             cxx,                /**< function coefficient */
   SCIP_Real             cxy,                /**< function coefficient */
   SCIP_Real             cyy,                /**< function coefficient */
   SCIP_Real             cx,                 /**< function coefficient */
   SCIP_Real             cy                  /**< function coefficient */
   )
{
   assert(lbx <= ubx);
   assert(lby <= uby);

   return 3.0 * MIN(cxxx * lbx, cxxx * ubx) + MIN(cxxy * lby, cxxy * uby) + cxx >= 0.0;
}

/* checks whether the 2 dimensional cubic function is edge convex, i.e., convex in one dimension if one fixes the other
   dimension */
static
SCIP_Bool isEdgeConvex(
   SCIP_Real             lbx,                /**< lower bound of x */
   SCIP_Real             ubx,                /**< upper bound of y */
   SCIP_Real             lby,                /**< lower bound of y */
   SCIP_Real             uby,                /**< upper bound of y */
   SCIP_Real             cxxx,               /**< function coefficient */
   SCIP_Real             cxxy,               /**< function coefficient */
   SCIP_Real             cxyy,               /**< function coefficient */
   SCIP_Real             cyyy,               /**< function coefficient */
   SCIP_Real             cxx,                /**< function coefficient */
   SCIP_Real             cxy,                /**< function coefficient */
   SCIP_Real             cyy,                /**< function coefficient */
   SCIP_Real             cx,                 /**< function coefficient */
   SCIP_Real             cy                  /**< function coefficient */
   )
{
   assert(lbx <= ubx);
   assert(lby <= uby);

   return isConvexInXOverBox(lbx, ubx, lby, uby, cxxx, cxxy, cxyy, cyyy, cxx, cxy, cyy, cx, cy) &&
          isConvexInXOverBox(lby, uby, lbx, ubx, cyyy, cxyy, cxxy, cxxx, cyy, cxy, cxx, cy, cx);
}

/* checks whether there is a point within the given bounds such that the hessian of the given 2 dimensional cubic
   function is indefinite */
static
SCIP_Bool isIndefiniteOverBox(
   SCIP_Real             lbx,                /**< lower bound of x */
   SCIP_Real             ubx,                /**< upper bound of y */
   SCIP_Real             lby,                /**< lower bound of y */
   SCIP_Real             uby,                /**< upper bound of y */
   SCIP_Real             cxxx,               /**< function coefficient */
   SCIP_Real             cxxy,               /**< function coefficient */
   SCIP_Real             cxyy,               /**< function coefficient */
   SCIP_Real             cyyy,               /**< function coefficient */
   SCIP_Real             cxx,                /**< function coefficient */
   SCIP_Real             cxy,                /**< function coefficient */
   SCIP_Real             cyy,                /**< function coefficient */
   SCIP_Real             cx,                 /**< function coefficient */
   SCIP_Real             cy                  /**< function coefficient */
   )
{
   assert(lbx <= ubx);
   assert(lby <= uby);
   assert(isEdgeConvex(lbx, ubx, lby, uby, cxxx, cxxy, cxyy, cyyy, cxx, cxy, cyy, cx, cy)); // otherwise we would need to solve the quadratic problem only over those parts of the box, where it is edge convex. Dont have a method for quadratic with linear constraints yet.

   return !(calcMin2dQuad(lbx, ubx, lby, uby,
         -12.0 * cxxx * cxyy + 4.0 * pow(cxxy, 2.0),
         -36.0 * cxxx * cyyy + 4.0 * cxxy * cxyy,
         -12.0 * cyyy * cxxy + 4.0 * pow(cxyy, 2.0),
         -12.0 * cxxx * cyy + 4.0 * cxxy * cxy - 4.0 * cxyy * cxx,
         -12.0 * cyyy * cxx + 4.0 * cxyy * cxy - 4.0 * cxxy * cyy,
         NULL, NULL)
      -4.0 * cxx * cyy + pow(cxy, 2.0) <= 0.0);
}

/* evaluate the given cubic for a given point */
static
SCIP_Real f_cube(
   SCIP_Real             x,                  /**< x value of point */
   SCIP_Real             y,                  /**< y value of point */
   SCIP_Real             cxxx,               /**< function coefficient */
   SCIP_Real             cxxy,               /**< function coefficient */
   SCIP_Real             cxyy,               /**< function coefficient */
   SCIP_Real             cyyy,               /**< function coefficient */
   SCIP_Real             cxx,                /**< function coefficient */
   SCIP_Real             cxy,                /**< function coefficient */
   SCIP_Real             cyy,                /**< function coefficient */
   SCIP_Real             cx,                 /**< function coefficient */
   SCIP_Real             cy                  /**< function coefficient */
   )
{
   return cxxx * pow(x, 3.0) + cxxy * pow(x, 2.0) * y + cxyy * x * pow(y, 2.0)
            + cyyy * pow(y, 3.0) + cxx * pow(x, 2.0) + cxy * x * y + cyy * pow(y, 2.0)
            + cx * x + cy * y;
}

/* get the x component of the gradient of given cubic function at a given point */
static
SCIP_Real GradientInX(
   SCIP_Real             x,                  /**< x value of point */
   SCIP_Real             y,                  /**< y value of point */
   SCIP_Real             cxxx,               /**< function coefficient */
   SCIP_Real             cxxy,               /**< function coefficient */
   SCIP_Real             cxyy,               /**< function coefficient */
   SCIP_Real             cyyy,               /**< function coefficient */
   SCIP_Real             cxx,                /**< function coefficient */
   SCIP_Real             cxy,                /**< function coefficient */
   SCIP_Real             cyy,                /**< function coefficient */
   SCIP_Real             cx,                 /**< function coefficient */
   SCIP_Real             cy                  /**< function coefficient */
   )
{
   return 3.0 * cxxx * x * x + 2.0 * cxxy * x * y + cxyy * y * y + 2.0 * cxx * x + cxy * y + cx;
}

/* get the y component of the gradient of given cubic function at a given point */
static
SCIP_Real GradientInY(
   SCIP_Real             x,                  /**< x value of point */
   SCIP_Real             y,                  /**< y value of point */
   SCIP_Real             cxxx,               /**< function coefficient */
   SCIP_Real             cxxy,               /**< function coefficient */
   SCIP_Real             cxyy,               /**< function coefficient */
   SCIP_Real             cyyy,               /**< function coefficient */
   SCIP_Real             cxx,                /**< function coefficient */
   SCIP_Real             cxy,                /**< function coefficient */
   SCIP_Real             cyy,                /**< function coefficient */
   SCIP_Real             cx,                 /**< function coefficient */
   SCIP_Real             cy                  /**< function coefficient */
   )
{
   return 3.0 * cyyy * y * y + 2.0 * cxyy * x * y + cxxy * x * x + 2.0 * cyy * y + cxy * x + cy;
}

/* check whether cutX * x + cutY * y + cutC <= f(x, y) for x, y in a box for a given cubic f */
static
SCIP_Bool CutIsFeasible(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_Real             lbx,                /**< lower bound of x */
   SCIP_Real             ubx,                /**< upper bound of y */
   SCIP_Real             lby,                /**< lower bound of y */
   SCIP_Real             uby,                /**< upper bound of y */
   SCIP_Real             cutX,               /**< cut coefficient */
   SCIP_Real             cutY,               /**< cut coefficient */
   SCIP_Real             cutC,               /**< cut coefficient */
   SCIP_Real             cxxx,               /**< function coefficient */
   SCIP_Real             cxxy,               /**< function coefficient */
   SCIP_Real             cxyy,               /**< function coefficient */
   SCIP_Real             cyyy,               /**< function coefficient */
   SCIP_Real             cxx,                /**< function coefficient */
   SCIP_Real             cxy,                /**< function coefficient */
   SCIP_Real             cyy,                /**< function coefficient */
   SCIP_Real             cx,                 /**< function coefficient */
   SCIP_Real             cy,                 /**< function coefficient */
   SCIP_Real             cc                  /**< function coefficient */
   )
{
   SCIP_Bool success;
   SCIP_Real min;
   SCIP_Real optX;
   SCIP_Real optY;

   min = calcMin2dCube(lbx, ubx, lby, uby, cxxx, cxxy, cxyy, cyyy, cxx, cxy, cyy, cx-cutX, cy-cutY, &optX, &optY, &success) + cc - cutC;

   return success && SCIPisFeasGE(scip, min, 0.0);
}

/* adds the cut varT >= coefX * varX + coefY * varY + coefZ * varZ if efficacious */
static
SCIP_RETCODE tryCut(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_SEPA*  	         sepa,               /**< separator that computed the cut */
   SCIP_SOL*             sol,                /**< primal solution that should be separated, or NULL for LP solution */
   SCIP_VAR*             varT,               /**< variable in the cut */
   SCIP_VAR*             varX,               /**< variable in the cut */
   SCIP_VAR*             varY,               /**< variable in the cut */
   SCIP_VAR*             varZ,               /**< variable in the cut */
   SCIP_Real             coefX,              /**< cut coefficient */
   SCIP_Real             coefY,              /**< cut coefficient */
   SCIP_Real             coefZ,              /**< cut coefficient */
   SCIP_RESULT*          result              /**< pointer to store the result of the separation call */
   )
{
   SCIP_ROW* row;
   SCIP_Bool infeasible;

   SCIP_CALL( SCIPcreateEmptyRowSepa(scip, &row, sepa, "cutSepa", 0.0, SCIPinfinity(scip), TRUE, FALSE, TRUE) );
   SCIP_CALL( SCIPcacheRowExtensions(scip, row) );

   SCIP_CALL( SCIPaddVarToRow(scip, row, varT, 1.0) );

   SCIP_CALL( SCIPaddVarToRow(scip, row, varX, -coefX) );
   SCIP_CALL( SCIPaddVarToRow(scip, row, varY, -coefY) );
   SCIP_CALL( SCIPaddVarToRow(scip, row, varZ, -coefZ) );

   SCIP_CALL( SCIPflushRowExtensions(scip, row) );

#ifdef SCIP_DEBUG
   SCIP_CALL( SCIPprintRow(scip, row, NULL) );
#endif

   if( SCIPisCutEfficacious(scip, sol, row) )
   {
      SCIP_CALL( SCIPaddRow(scip, row, TRUE, &infeasible) );
      if( infeasible )
         *result = SCIP_CUTOFF;
      else
         *result = SCIP_SEPARATED;
   }

   SCIP_CALL( SCIPreleaseRow(scip, &row) );

   return SCIP_OKAY;
}

/* compute cut, which lifts from the lower bound */
static
SCIP_RETCODE getCutFromLb(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_Real             y,                  /**< y value */
   SCIP_Real             lbX,                /**< variable bound */
   SCIP_Real             lbY,                /**< variable bound */
   SCIP_Real             ubY,                /**< variable bound */
   SCIP_Real             coefXXX,            /**< function coefficient */
   SCIP_Real             coefXXY,            /**< function coefficient */
   SCIP_Real             coefXYY,            /**< function coefficient */
   SCIP_Real             coefYYY,            /**< function coefficient */
   SCIP_Real             coefXX,             /**< function coefficient */
   SCIP_Real             coefXY,             /**< function coefficient */
   SCIP_Real             coefYY,             /**< function coefficient */
   SCIP_Real             coefX,              /**< function coefficient */
   SCIP_Real             coefY,              /**< function coefficient */
   SCIP_Real             coefC,              /**< function coefficient */
   SCIP_Real*            cutX,               /**< coefficient of computed cut */
   SCIP_Real*            cutY,               /**< coefficient of computed cut */
   SCIP_Real*            cutC                /**< coefficient of computed cut */
   )
{
   assert(scip != NULL);
   assert(cutX != NULL);
   assert(cutY != NULL);
   assert(cutC != NULL);

   if( !(SCIPisFeasLE(scip, lbY, y) && SCIPisFeasLE(scip, y, ubY)) )
   {
      printf("y does not belong to interval %f [%f %f]\n", y, lbY, ubY);
      return SCIP_ERROR;
   }

   if( GradientInX(lbX, lbY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) <=
      GradientInX(lbX, ubY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) )
   {
      *cutX = GradientInX(lbX, lbY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY);
   }
   else
   {
      *cutX = GradientInX(lbX, ubY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY);
   }

   *cutY = GradientInY(lbX, y, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY);

   *cutC = f_cube(lbX, y, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) + coefC;
   *cutC -= *cutX * lbX;
   *cutC -= *cutY * y;

   return SCIP_OKAY;
}

/* compute cut, which lifts from the lower bound */
static
SCIP_RETCODE getCutFromUb(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_Real             y,                  /**< y value */
   SCIP_Real             ubX,                /**< variable bound */
   SCIP_Real             lbY,                /**< variable bound */
   SCIP_Real             ubY,                /**< variable bound */
   SCIP_Real             coefXXX,            /**< function coefficient */
   SCIP_Real             coefXXY,            /**< function coefficient */
   SCIP_Real             coefXYY,            /**< function coefficient */
   SCIP_Real             coefYYY,            /**< function coefficient */
   SCIP_Real             coefXX,             /**< function coefficient */
   SCIP_Real             coefXY,             /**< function coefficient */
   SCIP_Real             coefYY,             /**< function coefficient */
   SCIP_Real             coefX,              /**< function coefficient */
   SCIP_Real             coefY,              /**< function coefficient */
   SCIP_Real             coefC,              /**< function coefficient */
   SCIP_Real*            cutX,               /**< coefficient of computed cut */
   SCIP_Real*            cutY,               /**< coefficient of computed cut */
   SCIP_Real*            cutC                /**< coefficient of computed cut */
   )
{
   assert(scip != NULL);
   assert(cutX != NULL);
   assert(cutY != NULL);
   assert(cutC != NULL);

   if( !(SCIPisFeasLE(scip, lbY, y) && SCIPisFeasLE(scip, y, ubY)) )
   {
      printf("y does not belong to interval %f [%f %f]\n", y, lbY, ubY);
      return SCIP_ERROR;
   }

   if( GradientInX(ubX, lbY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) >=
      GradientInX(ubX, ubY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) )
   {
      *cutX = GradientInX(ubX, lbY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY);
   }
   else
   {
      *cutX = GradientInX(ubX, ubY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY);
   }
   *cutY = GradientInY(ubX, y, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY);

   *cutC = f_cube(ubX, y, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) + coefC;
   *cutC -= *cutX * ubX;
   *cutC -= *cutY * y;

   return SCIP_OKAY;
}

/** adds the row f(x, y) >= col[0] + col[1] * (x-xx) + col[2] * (y-yy) to the given lpu*/
static
SCIP_RETCODE addLPIRow(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_LPI*             lpi,                /**< lp data structure */
   SCIP_Real             x,                  /**< points */
   SCIP_Real             y,                  /**< points */
   SCIP_Real             xx,                 /**< points */
   SCIP_Real             yy,                 /**< points */
   SCIP_Real             coefXXX,            /**< coefficient of f */
   SCIP_Real             coefXXY,            /**< coefficient of f */
   SCIP_Real             coefXYY,            /**< coefficient of f */
   SCIP_Real             coefYYY,            /**< coefficient of f */
   SCIP_Real             coefXX,             /**< coefficient of f */
   SCIP_Real             coefXY,             /**< coefficient of f */
   SCIP_Real             coefYY,             /**< coefficient of f */
   SCIP_Real             coefX,              /**< coefficient of f */
   SCIP_Real             coefY,              /**< coefficient of f */
   SCIP_Real             coefC               /**< coefficient of f */
   )
{
   SCIP_Real lb;
   SCIP_Real ub;
   int nnonz = 3;
   int beg = 0;
   SCIP_Real vals[3] = {1.0, x-xx, y-yy};
   int ind[3] = {0, 1, 2};

   lb = -SCIPinfinity(scip);
   ub = f_cube(x, y, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) + coefC;

   // lpiAddRows expects the coefficients to be nonzero
   if( SCIPisZero(scip, vals[2]) )
   {
      nnonz = 2;
   }
   if( SCIPisZero(scip, vals[1]) )
   {
      nnonz--;
      vals[1] = vals[2];
      ind[1] = ind[2];
   }

   SCIP_CALL( SCIPlpiAddRows(lpi, 1, &lb, &ub, NULL, nnonz, &beg, ind, vals) );
   //   SCIP_CALL( SCIPlpiWriteLP(lpi, "bla.lp") );
   return SCIP_OKAY;
}

/** tries out gradient cuts */
static
SCIP_RETCODE separateGradientCuts(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_SEPA*  	         sepa,               /**< separator that computes the cut */
   SCIP_SOL*             sol,                /**< primal solution that should be separated, or NULL for LP solution */
   SCIP_VAR*             varT,               /**< variable in the cut */
   SCIP_VAR*             varX,               /**< variable in the cut */
   SCIP_VAR*             varY,               /**< variable in the cut */
   SCIP_VAR*             varZ,               /**< variable in the cut */
   SCIP_Real             coefXXX,            /**< function coefficient */
   SCIP_Real             coefXXY,            /**< function coefficient */
   SCIP_Real             coefXYY,            /**< function coefficient */
   SCIP_Real             coefYYY,            /**< function coefficient */
   SCIP_Real             coefXX,             /**< function coefficient */
   SCIP_Real             coefXY,             /**< function coefficient */
   SCIP_Real             coefYY,             /**< function coefficient */
   SCIP_Real             coefX,              /**< function coefficient */
   SCIP_Real             coefY,              /**< function coefficient */
   SCIP_Real             coefC,              /**< function coefficient */
   SCIP_RESULT*          result              /**< pointer to store the result of the separation call */
   )
{
   SCIP_Real cutX;
   SCIP_Real cutY;
   SCIP_Real cutZ;
   SCIP_Real valX;
   SCIP_Real valY;
   SCIP_Real valZ;
   SCIP_Real lbX;
   SCIP_Real ubX;
   SCIP_Real lbY;
   SCIP_Real ubY;

   assert(scip != NULL);
   assert( varT != NULL );
   assert( varX != NULL );
   assert( varY != NULL );
   assert( varZ != NULL );

   valZ = SCIPgetSolVal(scip, sol, varZ);

   if( SCIPisFeasZero(scip, valZ) )
   {
      return SCIP_OKAY;
   }

   valX = SCIPgetSolVal(scip, sol, varX);
   valY = SCIPgetSolVal(scip, sol, varY);

   lbX = SCIPvarGetLbLocal(varX);
   ubX = SCIPvarGetUbLocal(varX);
   lbY = SCIPvarGetLbLocal(varY);
   ubY = SCIPvarGetUbLocal(varY);

   if( !SCIPisFeasEQ(scip, valZ, 1.0) )
   {
      valX = valX / valZ;
      valY = valY / valZ;
   }
   // sometimes the bounds are not tight, TODO implement separator lbX * z <= x<= ubX * z?
   valX = MAX(lbX, MIN(valX, ubX));
   valY = MAX(lbY, MIN(valY, ubY));

   cutX = GradientInX(valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY);
   cutY = GradientInY(valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY);
   cutZ = f_cube(valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) + coefC;
   cutZ -= cutX * valX;
   cutZ -= cutY * valY;

   if (!CutIsFeasible(scip, lbX, ubX, lbY, ubY, cutX, cutY, cutZ, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) )
   {
         return SCIP_OKAY;
   }


   SCIP_CALL( tryCut(scip, sepa, sol, varT, varX, varY, varZ, cutX, cutY, cutZ, result) );

   return SCIP_OKAY;
}

/** tries out cuts which lift from the boundary */
static
SCIP_RETCODE separateInteriorCut(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_SEPA*  	         sepa,               /**< separator that computes the cut */
   SCIP_SOL*             sol,                /**< primal solution that should be separated, or NULL for LP solution */
   SCIP_VAR*             varT,               /**< variable in the cut */
   SCIP_VAR*             varX,               /**< variable in the cut */
   SCIP_VAR*             varY,               /**< variable in the cut */
   SCIP_VAR*             varZ,               /**< variable in the cut */
   SCIP_Real             coefXXX,            /**< function coefficient */
   SCIP_Real             coefXXY,            /**< function coefficient */
   SCIP_Real             coefXYY,            /**< function coefficient */
   SCIP_Real             coefYYY,            /**< function coefficient */
   SCIP_Real             coefXX,             /**< function coefficient */
   SCIP_Real             coefXY,             /**< function coefficient */
   SCIP_Real             coefYY,             /**< function coefficient */
   SCIP_Real             coefX,              /**< function coefficient */
   SCIP_Real             coefY,              /**< function coefficient */
   SCIP_Real             coefC,              /**< function coefficient */
   SCIP_RESULT*          result              /**< pointer to store the result of the separation call */
   )
{
   char* name;
   SCIP_LPI* lpi;
   SCIP_Real obj;
   SCIP_Real lb;
   SCIP_Real ub;
   SCIP_Real valT;
   SCIP_Real valX;
   SCIP_Real valY;
   SCIP_Real valZ;
   SCIP_Real lbX;
   SCIP_Real ubX;
   SCIP_Real lbY;
   SCIP_Real ubY;
   SCIP_Real xopt;
   SCIP_Real yopt;
   SCIP_Bool addedCut;
   SCIP_Bool success;
   SCIP_Real coefsCut[3];
   SCIP_Real error;
   int i;

   assert(scip != NULL);

   assert( varT != NULL );
   assert( varX != NULL );
   assert( varY != NULL );
   assert( varZ != NULL );
   assert( result != NULL );

   valZ = SCIPgetSolVal(scip, sol, varZ);

   if( SCIPisFeasZero(scip, valZ) )
   {
      return SCIP_OKAY;
   }


   valX = SCIPgetSolVal(scip, sol, varX);
   valY = SCIPgetSolVal(scip, sol, varY);
   lbX = SCIPvarGetLbLocal(varX);
   ubX = SCIPvarGetUbLocal(varX);

   lbY = SCIPvarGetLbLocal(varY);
   ubY = SCIPvarGetUbLocal(varY);

   valT = SCIPgetSolVal(scip, sol, varT);

   if( !SCIPisFeasEQ(scip, valZ, 1.0) )
   {
      valX = valX / valZ;
      valY = valY / valZ;
   }
   // sometimes the bounds are not tight, TODO implement separator lbX * z <= x<= ubX * z?
   valX = MAX(lbX, MIN(valX, ubX));
   valY = MAX(lbY, MIN(valY, ubY));

   if( SCIPisEQ(scip, valX, lbX) || SCIPisEQ(scip, valX, ubX) || SCIPisEQ(scip, valY, lbY) || SCIPisEQ(scip, valY, ubY) )
   {
      return SCIP_OKAY;
   }

   if( SCIPisFeasGE(scip, valT, f_cube(valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) + coefC) )
   {
      return SCIP_OKAY;
   }

   SCIP_CALL( SCIPlpiCreate(&lpi, SCIPgetMessagehdlr(scip), "LP", SCIP_OBJSEN_MAXIMIZE) );

   /* add a dummy row which we later delete. this way we can add each column separately making it easier to name them */
   lb = 1.0;
   SCIP_CALL( SCIPlpiAddRows(lpi, 1, &lb, &lb, NULL, 0, NULL, NULL, NULL) );

   SCIP_CALL( SCIPallocBufferArray(scip, &name, SCIP_MAXSTRLEN) );

   lb = -SCIPinfinity(scip);
   ub = SCIPinfinity(scip);
   obj = 1.0;
   (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "c");
   SCIP_CALL( SCIPlpiAddCols(lpi, 1, &obj, &lb, &ub, &name, 0, NULL, NULL, NULL) );
   obj = 0.0;
   (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "coefX");
   SCIP_CALL( SCIPlpiAddCols(lpi, 1, &obj, &lb, &ub, &name, 0, NULL, NULL, NULL) );
   (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "coefY");
   SCIP_CALL( SCIPlpiAddCols(lpi, 1, &obj, &lb, &ub, &name, 0, NULL, NULL, NULL) );
   SCIP_CALL( SCIPlpiDelRows(lpi, 0, 0) );

   SCIPfreeBufferArray(scip, &name);


   /* add rows which correspond to the corners of the feasible box */
   SCIP_CALL( addLPIRow(scip, lpi, lbX, lbY, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );
   SCIP_CALL( addLPIRow(scip, lpi, lbX, ubY, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );
   SCIP_CALL( addLPIRow(scip, lpi, ubX, lbY, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );
   SCIP_CALL( addLPIRow(scip, lpi, ubX, ubY, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );
   /* also add rows which correspond to the minima of the function over the boundary */
   (void) calcMin1dCube(lbX, ubX, coefXXX, coefXXY * lbY + coefXX, coefXYY * pow(lbY, 2.0) + coefX + coefXY * lbY, &xopt);
   SCIP_CALL( addLPIRow(scip, lpi, xopt, lbY, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );
   (void) calcMin1dCube(lbX, ubX, coefXXX, coefXXY * ubY + coefXX, coefXYY * pow(ubY, 2.0) + coefX + coefXY * ubY, &xopt);
   SCIP_CALL( addLPIRow(scip, lpi, xopt, ubY, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );

   (void) calcMin1dCube(lbY, ubY, coefYYY, coefXYY * lbX + coefYY, coefXXY * pow(lbX, 2.0) + coefY + coefXY * lbX, &yopt);
   SCIP_CALL( addLPIRow(scip, lpi, lbX, yopt, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );
   (void) calcMin1dCube(lbY, ubY, coefYYY, coefXYY * ubX + coefYY, coefXXY * pow(ubX, 2.0) + coefY + coefXY * ubX, &yopt);
   SCIP_CALL( addLPIRow(scip, lpi, ubX, yopt, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );

   addedCut = TRUE;
   success = TRUE;
   i = 0;
   while( addedCut )
   {
      i++;
      //SCIP_CALL( SCIPlpiSolvePrimal(consdata->lpi_sep) );
      //SCIP_CALL( SCIPlpiWriteLP(consdata->lpi_sep, "sepa.lp") );
      SCIP_CALL( SCIPlpiSolveDual(lpi) );
      //SCIP_CALL( SCIPlpiSolveBarrier(consdata->lpi_sep, TRUE) );
      assert( SCIPlpiWasSolved(lpi) );
      if( !SCIPlpiIsPrimalFeasible(lpi) )
      {
         printf("primal infeasible\n");

         SCIP_CALL( SCIPlpiFree(&lpi) );
         return SCIP_OKAY;
      }
      if( !SCIPlpiIsOptimal(lpi) )
      {
         printf("not solved to optimality\n");
         return SCIP_ERROR;
      }
      SCIP_CALL( SCIPlpiGetSol(lpi, NULL, coefsCut, NULL, NULL, NULL) );

      //      printf("value %f\n", coefsCut[0]);
      if( SCIPisFeasGE(scip, valT, coefsCut[0]) || i > 100)
      {
         success = FALSE;
         break;
      }

      /* search for a point which is violated by the underestimate */
      error = calcMin2dCube(lbX, ubX, lbY, ubY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX - coefsCut[1], coefY - coefsCut[2], &xopt, &yopt, &success) + coefC - coefsCut[0] + coefsCut[1] * valX + coefsCut[2] * valY;

      if( !SCIPisFeasGE(scip, error, 0.0) )
      {
         SCIP_CALL( addLPIRow(scip, lpi, xopt, yopt, valX, valY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) );
      }
      else
      {
         addedCut = FALSE;
      }

   }
   if( success )
   {
      coefsCut[0] -= coefsCut[1] * valX;
      coefsCut[0] -= coefsCut[2] * valY;
      if (!CutIsFeasible(scip, lbX, ubX, lbY, ubY, coefsCut[1], coefsCut[2], coefsCut[0], coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) )
      {
         printf("cut not feasible\n");
         return SCIP_ERROR;
      }
      SCIP_CALL( tryCut(scip, sepa, sol, varT, varX, varY, varZ, coefsCut[1], coefsCut[2], coefsCut[0], result) );
   }
   SCIP_CALL( SCIPlpiFree(&lpi) );


   return SCIP_OKAY;
}

/** tries out cuts which lift from the boundary */
static
SCIP_RETCODE separateBorderCuts(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_SEPA*  	         sepa,               /**< separator that computes the cut */
   SCIP_SOL*             sol,                /**< primal solution that should be separated, or NULL for LP solution */
   SCIP_VAR*             varT,               /**< variable in the cut */
   SCIP_VAR*             varX,               /**< variable in the cut */
   SCIP_VAR*             varY,               /**< variable in the cut */
   SCIP_VAR*             varZ,               /**< variable in the cut */
   SCIP_Real             coefXXX,            /**< function coefficient */
   SCIP_Real             coefXXY,            /**< function coefficient */
   SCIP_Real             coefXYY,            /**< function coefficient */
   SCIP_Real             coefYYY,            /**< function coefficient */
   SCIP_Real             coefXX,             /**< function coefficient */
   SCIP_Real             coefXY,             /**< function coefficient */
   SCIP_Real             coefYY,             /**< function coefficient */
   SCIP_Real             coefX,              /**< function coefficient */
   SCIP_Real             coefY,              /**< function coefficient */
   SCIP_Real             coefC,              /**< function coefficient */
   SCIP_RESULT*          result              /**< pointer to store the result of the separation call */
   )
{
   SCIP_Real cutX;
   SCIP_Real cutY;
   SCIP_Real cutZ;
   SCIP_Real valX;
   SCIP_Real valY;
   SCIP_Real valZ;
   SCIP_Real lbX;
   SCIP_Real ubX;
   SCIP_Real lbY;
   SCIP_Real ubY;

   assert(scip != NULL);
   assert( varT != NULL );
   assert( varX != NULL );
   assert( varY != NULL );
   assert( varZ != NULL );
   assert( result != NULL );

   valZ = SCIPgetSolVal(scip, sol, varZ);

   if( SCIPisFeasZero(scip, valZ) )
   {
      return SCIP_OKAY;
   }


   valX = SCIPgetSolVal(scip, sol, varX);
   valY = SCIPgetSolVal(scip, sol, varY);

   lbX = SCIPvarGetLbLocal(varX);
   ubX = SCIPvarGetUbLocal(varX);
   lbY = SCIPvarGetLbLocal(varY);
   ubY = SCIPvarGetUbLocal(varY);

   if( !SCIPisFeasEQ(scip, valZ, 1.0) )
   {
      valX = valX / valZ;
      valY = valY / valZ;
   }
   // sometimes the bounds are not tight, TODO implement separator lbX * z <= x<= ubX * z?
   valX = MAX(lbX, MIN(valX, ubX));
   valY = MAX(lbY, MIN(valY, ubY));

   if( SCIPisEQ(scip, valX, lbX) )
   {
      SCIP_CALL( getCutFromLb(scip, valY, lbX, lbY, ubY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC, &cutX, &cutY, &cutZ) );

      if (!CutIsFeasible(scip, lbX, ubX, lbY, ubY, cutX, cutY, cutZ, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) )
      {
         return SCIP_ERROR;
      }


      SCIP_CALL( tryCut(scip, sepa, sol, varT, varX, varY, varZ, cutX, cutY, cutZ, result) );
      if( *result == SCIP_CUTOFF )
         return SCIP_OKAY;
   }
   if( SCIPisEQ(scip, valX, ubX) )
   {
      SCIP_CALL( getCutFromUb(scip, valY, ubX, lbY, ubY, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC, &cutX, &cutY, &cutZ) );

      if (!CutIsFeasible(scip, lbX, ubX, lbY, ubY, cutX, cutY, cutZ, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) )
      {
         return SCIP_ERROR;
      }

      SCIP_CALL( tryCut(scip, sepa, sol, varT, varX, varY, varZ, cutX, cutY, cutZ, result) );
      if( *result == SCIP_CUTOFF )
         return SCIP_OKAY;
   }
   if( SCIPisEQ(scip, valY, lbY) )
   {
      SCIP_CALL( getCutFromLb(scip, valX, lbY, lbX, ubX, coefYYY, coefXYY, coefXXY, coefXXX, coefYY, coefXY, coefXX, coefY, coefX, coefC, &cutY, &cutX, &cutZ) );

      if (!CutIsFeasible(scip, lbX, ubX, lbY, ubY, cutX, cutY, cutZ, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) )
      {
         return SCIP_ERROR;
      }

      SCIP_CALL( tryCut(scip, sepa, sol, varT, varX, varY, varZ, cutX, cutY, cutZ, result) );
      if( *result == SCIP_CUTOFF )
         return SCIP_OKAY;
   }
   if( SCIPisEQ(scip, valX, ubX) )
   {
      SCIP_CALL( getCutFromUb(scip, valX, ubY, lbX, ubX, coefYYY, coefXYY, coefXXY, coefXXX, coefYY, coefXY, coefXX, coefY, coefX, coefC, &cutY, &cutX, &cutZ) );

      if (!CutIsFeasible(scip, lbX, ubX, lbY, ubY, cutX, cutY, cutZ, coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY, coefC) )
      {
         return SCIP_ERROR;
      }

      SCIP_CALL( tryCut(scip, sepa, sol, varT, varX, varY, varZ, cutX, cutY, cutZ, result) );
      if( *result == SCIP_CUTOFF )
         return SCIP_OKAY;
   }

   return SCIP_OKAY;
}

/* separate all saved inequalities in the given separator for the given solution */
static
SCIP_RETCODE separateCuts(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_SEPA*            sepa,               /**< the cut separator itself */
   SCIP_SOL*             sol,                /**< primal solution that should be separated, or NULL for LP solution */
   SCIP_RESULT*          result              /**< pointer to store the result of the separation call */
   )
{  /*lint --e{715}*/
   SCIP_SEPADATA* sepadata;
   int i;

   assert(scip != NULL);
   assert(sepa != NULL);
   assert(*result == SCIP_DIDNOTRUN);

   /* get separator data */
   sepadata = SCIPsepaGetData(sepa);
   assert(sepadata != NULL);

   *result = SCIP_DIDNOTFIND;
   for( i = 0; i < sepadata->neqs; ++i )
   {
      SCIP_CALL( separateBorderCuts(scip, sepa, sol, sepadata->varsT[i], sepadata->varsX[i], sepadata->varsY[i], sepadata->varsZ[i],
            sepadata->coefsXXX[i],
            sepadata->coefsXXY[i],
            sepadata->coefsXYY[i],
            sepadata->coefsYYY[i],
            sepadata->coefsXX[i],
            sepadata->coefsXY[i],
            sepadata->coefsYY[i],
            sepadata->coefsX[i],
            sepadata->coefsY[i],
            sepadata->coefsC[i],
            result) );
      if( *result == SCIP_CUTOFF )
         return SCIP_OKAY;

      SCIP_CALL( separateGradientCuts(scip, sepa, sol, sepadata->varsT[i], sepadata->varsX[i], sepadata->varsY[i], sepadata->varsZ[i],
            sepadata->coefsXXX[i],
            sepadata->coefsXXY[i],
            sepadata->coefsXYY[i],
            sepadata->coefsYYY[i],
            sepadata->coefsXX[i],
            sepadata->coefsXY[i],
            sepadata->coefsYY[i],
            sepadata->coefsX[i],
            sepadata->coefsY[i],
            sepadata->coefsC[i],
            result) );
      if( *result == SCIP_CUTOFF )
         return SCIP_OKAY;

      SCIP_CALL( separateInteriorCut(scip, sepa, sol, sepadata->varsT[i], sepadata->varsX[i], sepadata->varsY[i], sepadata->varsZ[i],
            sepadata->coefsXXX[i],
            sepadata->coefsXXY[i],
            sepadata->coefsXYY[i],
            sepadata->coefsYYY[i],
            sepadata->coefsXX[i],
            sepadata->coefsXY[i],
            sepadata->coefsYY[i],
            sepadata->coefsX[i],
            sepadata->coefsY[i],
            sepadata->coefsC[i],
            result) );
      if( *result == SCIP_CUTOFF )
         return SCIP_OKAY;

   }

   return SCIP_OKAY;
}

/** adds variables and the coefficients of cubic polynom to separate to the separator */
SCIP_RETCODE SCIPSepaConvexCubeRegister(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_VAR*             varT,               /**< variable t */
   SCIP_VAR*             varX,               /**< variable x */
   SCIP_VAR*             varY,               /**< variable y */
   SCIP_VAR*             varZ,               /**< indicator variable*/
   SCIP_Real             coefXXX,            /**< coefficient of function */
   SCIP_Real             coefXXY,            /**< coefficient of function */
   SCIP_Real             coefXYY,            /**< coefficient of function */
   SCIP_Real             coefYYY,            /**< coefficient of function */
   SCIP_Real             coefXX,             /**< coefficient of function */
   SCIP_Real             coefXY,             /**< coefficient of function */
   SCIP_Real             coefYY,             /**< coefficient of function */
   SCIP_Real             coefX,              /**< coefficient of function */
   SCIP_Real             coefY,              /**< coefficient of function */
   SCIP_Real             coefC               /**< coefficient of function */
   )
{
   SCIP_SEPA* sepa;
   SCIP_SEPADATA* sepadata;

   assert(scip != NULL);

   sepa = SCIPfindSepa(scip, SEPA_NAME);

   if (sepa == NULL)
   {
      printf("no sepa found\n");
      return SCIP_ERROR;
   }

   sepadata = SCIPsepaGetData(sepa);
   if (sepadata == NULL)
   {
      printf("no sepadata found\n");
      return SCIP_ERROR;
   }
   // realloc memory if necessary
   if (sepadata->sizeeqs == sepadata->neqs+1 )
   {
      sepadata->sizeeqs = SCIPcalcMemGrowSize(scip, sepadata->sizeeqs+1);
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->varsT, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->varsX, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->varsY, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->varsZ, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsXXX, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsXXY, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsXYY, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsYYY, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsXX, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsXY, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsYY, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsX, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsY, sepadata->neqs+1, sepadata->sizeeqs) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &sepadata->coefsC, sepadata->neqs+1, sepadata->sizeeqs) );
   }

   //check the convexity properties
   if( !isEdgeConvex(SCIPvarGetLbGlobal(varX), SCIPvarGetUbGlobal(varX), SCIPvarGetLbGlobal(varY), SCIPvarGetUbGlobal(varY), coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) )
   {
      printf("have registry which is not edge convex\n");
      return SCIP_INVALIDDATA;
   }
   if( !isIndefiniteOverBox(SCIPvarGetLbGlobal(varX), SCIPvarGetUbGlobal(varX), SCIPvarGetLbGlobal(varY), SCIPvarGetUbGlobal(varY), coefXXX, coefXXY, coefXYY, coefYYY, coefXX, coefXY, coefYY, coefX, coefY) )
   {
      if( coefXXX == 0.0 && coefXXY == 0.0 && coefXYY == 0.0 && coefYYY == 0.0 && coefXX == 0.0 && coefXY == 0.0 && coefYY == 0.0 && coefX == 0.0 && coefY == 0.0 )
         return SCIP_OKAY;
      printf("have registry which is not indefinite inside the box\n");
      return SCIP_INVALIDDATA;
   }

   sepadata->varsT[sepadata->neqs] = varT;
   sepadata->varsX[sepadata->neqs] = varX;
   sepadata->varsY[sepadata->neqs] = varY;
   sepadata->varsZ[sepadata->neqs] = varZ;
   sepadata->coefsXXX[sepadata->neqs] = coefXXX;
   sepadata->coefsXXY[sepadata->neqs] = coefXXY;
   sepadata->coefsXYY[sepadata->neqs] = coefXYY;
   sepadata->coefsYYY[sepadata->neqs] = coefYYY;
   sepadata->coefsXX[sepadata->neqs] = coefXX;
   sepadata->coefsXY[sepadata->neqs] = coefXY;
   sepadata->coefsYY[sepadata->neqs] = coefYY;
   sepadata->coefsX[sepadata->neqs] = coefX;
   sepadata->coefsY[sepadata->neqs] = coefY;
   sepadata->coefsC[sepadata->neqs] = coefC;
   sepadata->neqs++;

   return SCIP_OKAY;
}



/*
 * Callback methods of separator
 */


/** destructor of separator to free user data (called when SCIP is exiting) */
static
SCIP_DECL_SEPAFREE(sepaFreeVexiter)
{  /*lint --e{715}*/

   SCIP_SEPADATA* sepadata;

   sepadata = SCIPsepaGetData(sepa);
   SCIPfreeBlockMemoryArray(scip, &sepadata->varsT, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->varsX, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->varsY, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->varsZ, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsXXX, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsXXY, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsXYY, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsYYY, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsXX, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsXY, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsYY, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsX, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsY, sepadata->sizeeqs);
   SCIPfreeBlockMemoryArray(scip, &sepadata->coefsC, sepadata->sizeeqs);

   SCIPfreeBlockMemory(scip, &sepadata);

   return SCIP_OKAY;
}


/** LP solution separation method of separator */
static
SCIP_DECL_SEPAEXECLP(sepaExeclpVexiter)
{  /*lint --e{715}*/
   *result = SCIP_DIDNOTRUN;

   /* only call separator, if we are not close to terminating */
   if( SCIPisStopped(scip) )
      return SCIP_OKAY;

   /* separate cuts on the LP solution */
   SCIP_CALL( separateCuts(scip, sepa, NULL, result) );

   return SCIP_OKAY;
}


/** arbitrary primal solution separation method of separator */
static
SCIP_DECL_SEPAEXECSOL(sepaExecsolVexiter)
{  /*lint --e{715}*/
   *result = SCIP_DIDNOTRUN;

   /* separate cuts on the given primal solution */
   SCIP_CALL( separateCuts(scip, sepa, sol, result) );

   return SCIP_OKAY;
}
/*
 * separator specific interface methods
 */

/** creates the separator and includes it in SCIP */
SCIP_RETCODE SCIPincludeSepaConvexCube(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_SEPADATA* sepadata;
   SCIP_SEPA* sepa;

   /* create separator data */
   sepadata = NULL;

   SCIP_CALL( SCIPallocBlockMemory(scip, &sepadata) );

   assert(sepadata != NULL);

   sepadata->neqs = 0;
   sepadata->sizeeqs = 16;

   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->varsT, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->varsX, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->varsY, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->varsZ, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsXXX, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsXXY, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsXYY, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsYYY, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsXX, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsXY, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsYY, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsX, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsY, sepadata->sizeeqs) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &sepadata->coefsC, sepadata->sizeeqs) );

   sepa = NULL;

   /* include separator */
   SCIP_CALL( SCIPincludeSepaBasic(scip, &sepa, SEPA_NAME, SEPA_DESC, SEPA_PRIORITY, SEPA_FREQ, SEPA_MAXBOUNDDIST,
         SEPA_USESSUBSCIP, SEPA_DELAY,
         sepaExeclpVexiter, sepaExecsolVexiter,
         sepadata) );

   assert(sepa != NULL);

   /* set non fundamental callbacks via setter functions */
   SCIP_CALL( SCIPsetSepaFree(scip, sepa, sepaFreeVexiter) );

   return SCIP_OKAY;
}

